/* dosxsymc.c, DOS eXtender functions using DPMI exclusively.
 * Not all functions were tested yet. Only those I needed very badly.
 *
 * By Alfons Hoogervorst / Alliterated Software Development, 1997.
 *
 * This one should be very portable. Also refer to the dosx.h header.
 */
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#if defined(PF_MSDOS_32)
#	include <mem.h>
#endif
#include <dos.h>
#include <dosx.h>

/* All functions return 0 if they are succesful. If an error occurred
 * they return a value != 0.
 *
 * I'm still confused by the DPMI specs. Some interrupt list seem to
 * differ on whether a particular DPMI version returns an error in
 * register AX... So, if an error occurs, and AX turns up empty,
 * my code returns __LINE__. Set the high bit to distinguish our
 * errors. */

#define DPMI_ERROR (int)(((DWORD)__LINE__) | 0x80000000UL)

int dosx_int(int intno, DOSX_REGS FAR* dr)
{
    struct SREGS s;
    union REGS r;

    memset(&r, 0, sizeof (r));
    segread(&s);

    WREGS(r).ax = dr->r.x.ax;
    WREGS(r).bx = dr->r.x.bx;
    WREGS(r).cx = dr->r.x.cx;
    WREGS(r).dx = dr->r.x.dx;
    WREGS(r).di = dr->r.x.di;
    WREGS(r).si = dr->r.x.si;

    s.es   = dr->s.es;
    s.ds   = dr->s.ds;

    INT86X(intno, &r, &r, &s);

    dr->r.x.ax = WREGS(r).ax;
    dr->r.x.bx = WREGS(r).bx;
    dr->r.x.cx = WREGS(r).cx;
    dr->r.x.dx = WREGS(r).dx;
    dr->r.x.di = WREGS(r).di;
    dr->r.x.si = WREGS(r).si;

    dr->flags  = CFLAG(r);
    dr->s.es   = s.es;
    dr->s.ds   = s.ds;

    return 0;
}

/***/

#if defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
void a11_vm_idle(void);
#pragma aux a11_vm_idle = \
   "mov ax,01680h" \
   "int 2fh"
#endif /* __WATCOMC__, AHX_COMPILER_NATIVE */

void
dosx_vm_idle(void)
{
#if (defined(__SC__) || defined(__BORLANDC__)) && defined(AHX_COMPILER_NATIVE)
    __asm mov ax,0x1680
    __asm int 0x2f
#elif defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
    a11_vm_idle();
#else
    union REGS r;
    WREGS(r).ax = 0x1680;
    INT86(0x2f, &r, &r);
#endif
}

/* The next stuff only valid for extended dos */

#if defined(PF_MSDOS_32)

int
dosx_dos_alloc(LPVOID_X FAR* news, unsigned siz)
{
    union REGS r;
    if ( news == NULL ) return DPMI_ERROR;
    WREGS(r).ax  = 0x100;
    EREGS(r).ebx = (siz + 15) / 16;
    INT86(0x31, &r, &r);
    news->pm.sel = WREGS(r).dx;
    news->pm.off = 0;
    news->rm.seg = WREGS(r).ax;
    news->rm.off = 0;
    if ( CFLAG(r) & 1 )
    {
        news->pm.sel = WREGS(r).bx;
        news->valid = 0;
    }
    else
    {
        news->valid = 1;
    }
    return news->valid ? 0 : news->pm.sel == 0 ? DPMI_ERROR : news->pm.sel;
}

int
dosx_dos_free(LPVOID_X old)
{
    union REGS r;
    if ( !old.valid )
        return 0x8022; /* Invalid selector */
    EREGS(r).eax = 0x0101;
    WREGS(r).dx = old.pm.sel;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

unsigned
dosx_selector_inc(void)
{
    union REGS r;
    WREGS(r).ax = 0x3;
    INT86(0x31, &r, &r);
    return WREGS(r).ax;
}

int
dosx_alloc_selector(SELECTOR FAR* sel, unsigned count)
{
    union REGS r;
    if ( sel == 0 ) return DPMI_ERROR;
    EREGS(r).eax = 0;
    EREGS(r).ecx = count;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? (*sel = WREGS(r).ax), 0
           : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_free_selector(SELECTOR sel)
{
    union REGS r;
    EREGS(r).eax = 1;
    WREGS(r).bx = sel;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_alloc_alias(SELECTOR FAR* alias, SELECTOR copy)
{
    union REGS r;
    if ( alias == NULL ) return DPMI_ERROR;
    WREGS(r).ax = 0xa;
    WREGS(r).bx = copy;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? (*alias = WREGS(r).ax), 0
           : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}


SELECTOR
dosx_current_ds(void)
{
#if (defined(__SC__) || defined(__BORLANDC__)) && defined(AHX_COMPILER_NATIVE)
    SELECTOR res;
    __asm mov res,ds
    return res;
#else
    struct SREGS s;
    segread(&s);
    return s.ds;
#endif
}

SELECTOR
dosx_current_cs(void)
{
#if (defined(__SC__) || defined(__BORLANDC__)) && defined(AHX_COMPILER_NATIVE)
    SELECTOR res;
    __asm mov res,cs
    return res;
#else
    struct SREGS s;
    segread(&s);
    return s.cs;
#endif
}

SELECTOR
dosx_current_ss(void)
{
#if (defined(__SC__) || defined(__BORLANDC__)) && defined(AHX_COMPILER_NATIVE)
    SELECTOR res;
    __asm mov res,ss
    return res;
#else
    struct SREGS s;
    segread(&s);
    return s.ss;
#endif
}

int
dosx_get_selector_base(SELECTOR sel, DWORD FAR* lin)
{
    union REGS r;
    WREGS(r).ax = 0x6;
    WREGS(r).bx = sel;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? (*lin = MAKELONG(WREGS(r).dx, WREGS(r).cx)), 0
           : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_set_selector_base(SELECTOR sel, DWORD lin)
{
    union REGS r;
    if ( sel == 0 ) return 0x8022; /* Invalid selector */
    WREGS(r).ax = 0x7;
    WREGS(r).bx = sel;
    WREGS(r).cx = lin >> 16;
    WREGS(r).dx = lin & 0xffff;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_get_descriptor(SELECTOR sel, DESCRIPTOR FAR* des)
{
    union REGS      r;
    struct SREGS    s;
    if ( sel == 0 ) return 0x8022;
    segread(&s);
    s.es = FP_SEG(des);
    WREGS(r).ax = 0xb;
    WREGS(r).bx = sel;
    EREGS(r).edi = (unsigned long)des;
    INT86X(0x31, &r, &r, &s);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_set_descriptor(SELECTOR sel, CONST DESCRIPTOR FAR* des)
{
    union REGS      r;
    struct SREGS    s;
    if ( sel == 0 ) return 0x8022;
    segread(&s);
    s.es = FP_SEG((LPVOID)des);
    WREGS(r).ax = 0xc;
    WREGS(r).bx = sel;
    EREGS(r).edi = (unsigned long)(LPCVOID)des;
    INT86X(0x31, &r, &r, &s);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_set_selector_limit(SELECTOR sel, DWORD lim)
{
    union REGS r;
    WREGS(r).ax = 0x08;
    WREGS(r).bx = sel;
    WREGS(r).dx = lim & 0xffff;
    WREGS(r).cx = lim >> 16;
    INT86(0x31, &r, &r);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_get_selector_limit(SELECTOR sel, DWORD FAR* lim)
{
    int result = 0;
    DWORD limit = 0;
    DESCRIPTOR des;
    if ( lim == NULL || sel == 0 ) return DPMI_ERROR;
    result = dosx_get_descriptor(sel, &des);
    assert(result == 0);
    if ( result == 0 )
    {
        limit = ((DWORD)(des.d.limitrts_hi & 0x0f) << 16UL) +
            (DWORD)des.d.limit_lo;
        if ( des.d.limitrts_hi & 0x80 ) limit *= 4096;
    }
    *lim = limit;
    return result;
}

int
dosx_call_far_real(LPVOID_RM proc, DOSX_REGS FAR* dr)
{
    struct SREGS s;
    union REGS   r;
    if ( dr == NULL || (proc.seg == 0 && proc.off == 0) )
        return DPMI_ERROR;
    segread(&s);
    memset(&r, 0, sizeof (r));
    dr->s.cs = proc.seg;
    dr->s.ip = proc.off;
    WREGS(r).ax = 0x301;
    WREGS(r).bx = 0;
    WREGS(r).cx = 0;
    s.es = FP_SEG(dr);
    EREGS(r).edi = (unsigned long)dr;
    INT86X(0x31, &r, &r, &s);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int
dosx_int_real(int intno, DOSX_REGS FAR* dr)
{
    struct SREGS s;
    union  REGS  r;
    if ( dr == NULL ) return DPMI_ERROR;
    segread(&s);
    s.es = FP_SEG(dr);
    WREGS(r).ax = 0x300;
    WREGS(r).bx = intno;
    WREGS(r).cx = 0;
    EREGS(r).edi = (unsigned long)dr;
    INT86X(0x31, &r, &r, &s);
    return !(CFLAG(r) & 1) ? 0 : WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

void
dosx_far_near_copy(LPVOID_PM dst, LPCVOID src, unsigned count)
{
#if defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
    _fmemcpy(MK_FP(dst.sel, dst.off),
        MK_FP(FP_SEG(src), FP_OFF(src)),
        count);
#else
/* Should support __far (48 bit FWORD) pointers */
    typedef BYTE __far* FARCBYTE;
    FARCBYTE pdst = MK_FP(dst.sel, dst.off),
             psrc = MK_FP(FP_SEG((LPVOID)src), FP_OFF((LPVOID)src));
    while ( count )
    {
        *pdst++ = *psrc++; count--;
    }
#endif
}

void
dosx_near_far_copy(LPVOID dst, LPVOID_PM src, unsigned count)
{
#if defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
    _fmemcpy(MK_FP(FP_SEG(dst), FP_OFF(dst)),
        MK_FP(src.sel, src.off), count);
#else
/* Should support __far (48 bit FWORD) pointers */
    typedef BYTE __far* FARCBYTE;
    FARCBYTE pdst = MK_FP(FP_SEG(dst), FP_OFF(dst)),
             psrc = MK_FP(src.sel, src.off);
    while ( count )
    {
        *pdst++ = *psrc++;
        count--;
    }
#endif
}

unsigned
dosx_far_strlen(LPVOID_PM str)
{
#if defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
    return _fstrlen(MK_FP(str.sel, str.off));
#else
/* Should support __far (48 bit FWORD) pointers */
    typedef BYTE __far* FARCBYTE;
    FARCBYTE pstr = MK_FP(str.sel, str.off);
    unsigned count = 0;
    while ( *pstr++ ) count++;
    return count;
#endif
}

void
dosx_far_set(LPVOID_PM ptr, int value, unsigned size)
{
#if defined(__WATCOMC__) && defined(AHX_COMPILER_NATIVE)
    _fmemset(MK_FP(dst.sel, dst.off), value, size);
#else
/* Should support __far (48 bit FWORD) pointers */
    typedef BYTE __far* FARCBYTE;
    FARCBYTE sptr  = MK_FP(ptr.sel, ptr.off);
    while ( size )
    {
        *sptr++ = (BYTE)value;
        size--;
    }
#endif
}

LPVOID_PM
dosx_far_address_of(LPCVOID p)
{
    LPVOID_PM result;
    result.sel = FP_SEG((LPVOID)p);
    result.off = FP_OFF((LPVOID)p);
    return result;
}

DWORD
dosx_linear_address_of(LPCVOID p)
{
    DWORD base;
    dosx_get_selector_base(FP_SEG((LPVOID)p), &base);
    return base + FP_OFF((LPVOID)p);
}

/* dosx_map_linear() - Maps a linear address to a SEL:OFF32 pointer (FWORD).
 * The returned LPVOID_PM should be freed with dosx_unmap_linear() (or
 * by freeing the selector in the structure) */
int
dosx_map_linear(LPVOID_PM FAR* res_ptr, DWORD linear)
{
    LPVOID_PM p;
    int result;
    if ( res_ptr == NULL ) return DPMI_ERROR;
/* FIXME: Assuming ds has an OK limit */
    result = dosx_alloc_alias(&p.sel, dosx_current_ds());
    assert(result);
    result = dosx_set_selector_base(p.sel, linear);
    assert(result);
    if ( result != 0 )
    {
        dosx_free_selector(p.sel);
        return DPMI_ERROR;
    }
    p.off = 0;
    *res_ptr = p;
    return 0;
}

int
dosx_unmap_linear(LPVOID_PM p)
{
    return dosx_free_selector(p.sel);
}

/* dosx_map_addr16() - Phoney, phoney, phoney. Equivalent to dosx_map_linear(),
 * but returns a 16 bit protected mode address. (But in a handy form.) */
DWORD
dosx_map_addr16(LPCVOID address)
{
    LPVOID_PM pm_addr;
    int result = dosx_map_linear(&pm_addr, dosx_linear_address_of(address));
    if ( result == 0 )
        return MAKELONG(pm_addr.off, pm_addr.sel);
    else return 0;
}

int
dosx_unmap_addr16(DWORD dw)
{
    return dosx_free_selector(HIWORD(dw));
}

/* dosx_make_selector16() - Twiggles a 32 bit selector in a 16 bit one.
 * Will affect instructions in the associated segment. */
int
dosx_make_selector16(SELECTOR s)
{
    int result;
    DESCRIPTOR four_seasons;
    result = dosx_get_descriptor(s, &four_seasons);
    assert(result);
    if ( result != 0 ) return DPMI_ERROR;
/* 2nd bit of 1st nibble of 6th byte of our pizza */
/* (For hungry programmers: 6th bit of 6th byte) */
    four_seasons.b[6] &= ~0x40;
    result = dosx_set_descriptor(s, &four_seasons);
    assert(result);
    return result;
}

/* dosx_presto_chango_selector() - Data selector -> code selector,
 * code selector -> data selector. Tribute to nifty Windows 3.x function,
 * which had to be exported as "ChangeSelector", but for some reason
 * was exported as "PrestoChangoSelector". */
int
dosx_presto_chango_selector(SELECTOR s)
{
    int result;
    DESCRIPTOR four_seasons;
    result = dosx_get_descriptor(s, &four_seasons);
    assert(result == 0);
    if ( result != 0 ) return DPMI_ERROR;
/* Flip bit 3 of type bits in byte 5:
 *
 *       7 65 4 321 0       A = Accessed
 *      +-+--+-+---+-+    TYP = Segment type
 *      |P|PL|S|TYP|A|  (D)PL = Descriptor Privilege Level
 *      +-+--+-+---+-+      P = Present
 *
 * For your information the possible values of TYP:
 *
 *  000 - read only
 *  001 - read/write
 *  010 - read-only, expand down
 *  011 - read-write, expand down
 *  100 - execute-only
 *  101 - execute/read
 *  110 - execute-only, conforming
 *  111 - execute/read-only, conforming
 */
    four_seasons.b[5] ^= 0x8; /* 1000b */
    result = dosx_set_descriptor(s, &four_seasons);
    assert(result == 0);
    return result;
}

int dosx_alloc_copy(SELECTOR FAR* dst, SELECTOR src)
{
    int result;
    SELECTOR s;
    DESCRIPTOR four_seasons;
    if ( dst == 0 ) return DPMI_ERROR;
    result = dosx_get_descriptor(src, &four_seasons);
    assert(0 == result);
    if ( result != 0 ) return DPMI_ERROR;
    result = dosx_alloc_alias(&s, src);
    assert(0 == result);
    if ( result != 0 ) return DPMI_ERROR;
/* FIXME: clean up this one. :) */
    if ( 0 != (result = dosx_set_descriptor(s, &four_seasons)) )
    {
        dosx_free_selector(s);
        return DPMI_ERROR;
    }
    *dst = s;
    return 0;
}

int dosx_lock_memory(LPVOID_PM address, unsigned size)
{
    union REGS r;
    DWORD      d;

    dosx_get_selector_base(address.sel, &d);
    d += address.off;
    memset(&r, 0, sizeof (r));
    WREGS(r).ax = 0x0600;
    WREGS(r).bx = d >> 16UL;
    WREGS(r).cx = d & 0xffffUL;
    WREGS(r).si = size >> 16UL;
    WREGS(r).di = size & 0xffffUL;
    INT86(0x31, &r, &r);
    return CFLAG(r) & 1 == 0 ? 0 :
           WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

int dosx_unlock_memory(LPVOID_PM address, unsigned size)
{
    union REGS r;
    DWORD      d;

    dosx_get_selector_base(address.sel, &d);
    d += address.off;
    memset(&r, 0, sizeof (r));
    WREGS(r).ax = 0x0601;
    WREGS(r).bx = d >> 16UL;
    WREGS(r).cx = d & 0xffffUL;
    WREGS(r).si = size >> 16UL;
    WREGS(r).di = size & 0xffffUL;
    INT86(0x31, &r, &r);
    return CFLAG(r) & 1 == 0 ? 0 :
           WREGS(r).ax == 0 ? DPMI_ERROR : WREGS(r).ax;
}

#endif /* PF_MSDOS_32 */
